/****************************************************** INCLUDES *****/

#define LITTLE_ENDIAN	1
#define BIG_ENDIAN	2
#define BYTE_ORDER	LITTLE_ENDIAN

#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <limits.h>
#include <signal.h>
#include <sys/psinfo.h>
#include <sys/sys_msg.h>
#include <sys/kernel.h>
#include <sys/sendmx.h>
#include <sys/prfx.h>
#include <sys/sched.h>
#include <sys/io_msg.h>
#include <sys/fd.h>
#include <time.h>
#include <sys/dev_msg.h>
#include <sys/proxy.h>
#include <sys/select.h>
#include <sys/socket.h>
#include <sys/select.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <sys/ioctl.h>
#include <sys/trace.h>
#include <sys/tracecod.h>
#include <fcntl.h>

#define DEBUG

#define _TRACE_INFLEET _TRACE_TEMPORARY

#define _TRACE_BROKEN	(_TRACE_INFLEET|1)
#define _TRACE_IN_CONN	(_TRACE_INFLEET|2)
#define _TRACE_OUT_CONN	(_TRACE_INFLEET|3)
#define _TRACE_DEATH	(_TRACE_INFLEET|4)

/************************************************************* STRUCTS ****/

union {
    msg_t               type;
    msg_t               status;

    struct _sysmsg_hdr  sysmsg;
    struct _sysmsg_hdr_reply version_reply;

    struct _io_open     open;
    struct _io_open_reply open_reply;
    struct _io_close    close;
    struct _io_close_reply close_reply;

    struct _io_read     read;
    struct _io_read_reply read_reply;
    struct _io_write    write;
    struct _io_write_reply write_reply;
    char                buf[2000];          /* make sure room for write */

    struct _io_fstat    fstat;
    struct _io_fstat_reply fstat_reply;

    struct _dev_tcflush         dev_tcflush;
    struct _dev_tcflush_reply   dev_tcflush_reply;
    struct _dev_mode            dev_mode;
    struct _dev_mode_reply      dev_mode_reply;
    struct _dev_tcgetattr       dev_tcgetattr;
    struct _dev_tcgetattr_reply dev_tcgetattr_reply;
    struct _dev_tcsetattr       dev_tcsetattr;
    struct _dev_tcsetattr_reply dev_tcsetattr_reply;
    struct _dev_state           dev_state;
    struct _dev_state_reply     dev_state_reply;
    struct _dev_osize           dev_osize;
    struct _dev_osize_reply     dev_osize_reply;
    struct _dev_read            dev_read;
    struct _dev_read_reply      dev_read_reply;
    struct _dev_arm             dev_arm;
    struct _dev_arm_reply       dev_arm_reply;
    struct _dev_tcsetct         dev_tcsetct;
    struct _dev_tcsetct_reply   dev_tcsetct_reply;
} msg;

struct _sysmsg_version_reply version_stamp = {
    "infleet",
    __DATE__,
    421,
    'X',
    0
};

/* 
 *  Net.fd encapsulates the ethernet packet as follows
 */
#define SYNC_LEN    4
#define SYNC_0      0x01
#define SYNC_1      0x5A
#define SYNC_2      0x6B
#define SYNC_3      0x7C

#define COMP_NONE   0

struct fd_hdr_struct {
    char                    sync[SYNC_LEN];
    short unsigned int      len;    /* of everything after fd_hdr */
    short unsigned int      comp_type;  /* for types of compression in future */
    unsigned long           fd_hdr_crc;
};

#define MAX_DATA	2048

struct data_info {
    pid_t		proxy;
    unsigned		used;
    unsigned		min;
    unsigned		error;
    char		buff[MAX_DATA];
};

struct connection_info {
    int			h;
    struct sockaddr_in 	socket;
    struct data_info	in;
    struct data_info	out;
    enum { UNCREATED, UNCONNECTED, WAITING, CONNECTED } 	state;
};

#define LEFT( b ) ((sizeof((b).buff)-(b).used))

struct _psinfo3         psdata3;
struct _io_fstat_reply  fstat_reply;
struct _fd_ctrl         *fd_ctrl;
struct connection_info	conn;
unsigned		max_write = MAX_DATA;
int			report_fd = 2;
int    timeout_value  = 10;

void terminate( int signal )
{
    signal = signal;
    close( conn.h );
    exit( 0 );
}

void broken( int signal )
{
    /* other side of the line has gone bye-bye */
    signal = signal;
    if( conn.in.error == EOK ) {
	shutdown( conn.h, 2 );
	conn.state = UNCREATED;
	conn.in.error = EIO;
	conn.out.error = EIO;
	if( conn.in.proxy != 0 ) {
	    Trigger( conn.in.proxy );
	    conn.in.proxy = 0;
	}
	if( conn.out.proxy != 0 ) {
	    Trigger( conn.out.proxy );
	    conn.out.proxy = 0;
	}
	Trace0b( _TRACE_BROKEN, _TRACE_EXPECTED, conn.socket.sin_len,
			&conn.socket.sin_addr );
    }
}

void report_error( char *who )
{
    unsigned len;

    len = sprintf( conn.out.buff, "infleet[%s] %s - %s\n",
    	inet_ntoa( conn.socket.sin_addr ), who, strerror( errno ) );
    write( report_fd, conn.out.buff, len );
    conn.out.buff[len] = '\0';
    Trace0b( _TRACE_DEATH, _TRACE_SEVERE, len, conn.out.buff );
    exit( EXIT_FAILURE );
}

/************************************************************* MAIN ******/

#ifdef __USAGE
%C	[-o max] [-p port] [-P priority] [-r root] host

Where:
-o max		Set maximum size to output atomically.
-p port		Set the TCP port number to use. Default is the value
		specified by the "infleet" service in /etc/services.
-P priority	Priority to run at. Default is 10.
-r root		Set the root of the filename the program adopts.
		Default is "/Inet".
-t select timeout. Default is 10 seconds. Set to 0 to poll.
host		Internet host name or network number to talk to. This
		string is also used as the basename of the filename the
		program adopts.
#endif


main(int argc, char *argv[])
{
    int                 devno;
    int			c;
    struct hostent 	*hp;
    struct servent	*sp;
    char		*root = "/Inet";
    int			prio = 10;
    int			forking = 0;
    int			port = 0;
    int			listen_socket;
    struct sockaddr_in	listen_data;

    while ((c=getopt(argc,argv,"do:P:p:r:t:")) != -1) {
        switch (c) {
	case 'd':
	    forking = -1; /* debugging */
	    break;
	case 'o':
            max_write = strtoul(optarg,NULL,0);
	    break;
	case 'P':
	    prio = strtoul( optarg, NULL, 0 );
	    break;
    case 'p':
        port = htonl( strtoul(optarg,NULL,0) );
        break;
	case 'r':
	    root = optarg;
	    break;
	case 't':
	    timeout_value = atoi(optarg);
	    break;
	default:
	    //ill option
	    exit( EXIT_FAILURE );
	}
    }
    if( max_write < 16 ) max_write = 16;
    if( port == 0 ) {
	sp = getservbyname( "infleet", "tcp" );
	if( sp == NULL ) {
	    fprintf( stderr, "infleet service not found and -p not used\n" );
	    exit( EXIT_FAILURE );
	}
	port = sp->s_port;
    }
    if( optind == 0 || argv[optind] == NULL ) {
	fprintf( stderr, "no host specified\n" );
	exit( EXIT_FAILURE );
    }
    hp = gethostbyname( argv[optind] );
    if( hp == 0 ) {
	fprintf( stderr, "'%s' - no such host\n", argv[optind] );
	exit( EXIT_FAILURE );
    }
    memcpy( &conn.socket.sin_addr, hp->h_addr, hp->h_length );
    conn.socket.sin_len = hp->h_length;
    conn.socket.sin_family = AF_INET;
    conn.socket.sin_port = port;

    if (qnx_pflags(_PPF_SERVER, _PPF_SERVER, NULL, NULL)) {
	report_error("pflags");
    }

    if( signal(SIGINT, SIG_IGN) != SIG_IGN && forking == 0 ) {
	pid_t	child;
	
	child = fork();
	if( child == -1 ) {
	    report_error( "fork" );
	}
	if( child != 0 ) {
	    /* wait for initialization to finish */
	    Receive( child, 0, 0 );
	    Reply( child, 0, 0 );
	    exit( 0 ); /* let's not keep the shell waiting */
	}
	setsid();
	forking = 1;
    }
/*
    close(0);
    close(1);
*/

    if ((devno = qnx_device_attach()) == -1) {
	report_error("device_attach");
    }
    fstat_reply.status        = EOK;
    fstat_reply.stat.st_ino   = 1;
    fstat_reply.stat.st_rdev  = (dev_t) ((getnid() << 16L) + (devno << 10L)) + 1;
    fstat_reply.stat.st_mode  = S_IFREG | S_IREAD | S_IWRITE;
    fstat_reply.stat.st_size  = 1500;
    fstat_reply.stat.st_nlink = 1;
    fstat_reply.stat.st_ftime =
     fstat_reply.stat.st_mtime =
     fstat_reply.stat.st_atime =
     fstat_reply.stat.st_ctime = 0;
    if (!(fd_ctrl = __init_fd(getpid()))) {
	report_error("__init_fd");
    }
    /* steal the use of the output buffer for a sec. */
    if( strcmp( root, "" ) == 0 ) root = ".";
    strcpy( conn.out.buff, root );
    if( strcmp( root, "/" ) != 0 ) strcat( conn.out.buff, "/" );
    strcat( conn.out.buff, argv[optind] );
    if (qnx_prefix_attach( conn.out.buff, NULL, 0)) {
	report_error("prefix_attach");
    }

    qnx_scheduler(0, 0, SCHED_RR, prio, 1);

    memset( &listen_data, 0, sizeof( listen_data ) );
    listen_data.sin_family = AF_INET;
    listen_data.sin_addr.s_addr = INADDR_ANY;
    listen_data.sin_port = port;
    listen_socket = socket( AF_INET, SOCK_STREAM, 0 );
    if( listen_socket < 0 ) {
	report_error( "listen socket" );
    }
    if( bind( listen_socket, (struct sockaddr *)&listen_data,
	      sizeof( listen_data ) ) ) {
	report_error( "bind" );
    }
    if( listen( listen_socket, 5 ) < 0 ) {
	report_error( "listen" );
    }

    (void) signal(SIGHUP, SIG_IGN);
    signal( SIGTERM, &terminate );
    signal( SIGPWR, &terminate );
    signal( SIGPIPE, &broken );

    if( forking > 0 ) Send( getppid(), 0, 0, 0, 0 );

    /* It's OK if this fails.... */
    report_fd = open( "/dev/console", O_WRONLY );

    ProcessLoop( listen_socket );
}


static void ProcessLoop( int listen_socket )
{
    fd_set		rfd;
    fd_set		wfd;
    int			len;
    int			status;
    struct timeval	timeout;

    conn.h = -1;
    FD_ZERO( &rfd );
    for( ;; ) {
	timeout.tv_sec =  timeout_value;     /* 10 was the default */
	timeout.tv_usec = 0;
	/* somebody wants to connect */
	if( FD_ISSET( listen_socket, &rfd ) ) {
	    int			new;
	    struct sockaddr_in 	peer;
	    int			peer_len = sizeof( peer );
	    struct sockaddr_in  mine;
	    int			mine_len = sizeof( mine );

	    /* in case we refuse the connection */
	    FD_CLR( listen_socket, &rfd );
	    new = accept( listen_socket, (struct sockaddr * )&peer, &peer_len );
	    if( new < 0 ) {
		report_error( "accept" );
	    }
	    switch( conn.state ) {
        case UNCREATED:
	    case UNCONNECTED:
	        break;
	    case WAITING:
	        /*
		    We've initiated a connection as well. Flip a coin as
		    to which one gets used.
		     */
		if( getsockname( new, (struct sockaddr *)&mine, &mine_len ) < 0 ) {
		    report_error( "getsockname" );
		}
		if( memcmp( &mine, &peer, min( mine_len, peer_len ) ) < 0 ) break;
		/* fall through */
	    case CONNECTED:
	        /* refuse the connection */
	        close( new );
		continue;
	    }
	    /* accept the connection */
	    close( conn.h );
	    conn.h = new;
	    nonblock( conn.h );
	    config_socket( conn.h );
	    conn.state = CONNECTED;
	    Trace0b( _TRACE_IN_CONN, _TRACE_TESTING,
			    conn.socket.sin_len,
			    &conn.socket.sin_addr );
	}
	FD_ZERO( &rfd );
	FD_ZERO( &wfd );
	switch( conn.state ) {
	case UNCREATED:
	    conn.in.used = 0;
	    conn.out.used = 0;
	    close( conn.h );
	    conn.h = socket(AF_INET, SOCK_STREAM, 0);
	    if( conn.h < 0 ) {
		report_error( "connection socket" );
	    }
	    nonblock( conn.h );
	    conn.state = UNCONNECTED;
	    /* fall through */
	case UNCONNECTED:
	    /* don't bother attempting to connect until there's data
	       to send
	    */
	    if( conn.out.used == 0 ) break;
	    if( connect( conn.h, (struct sockaddr*)&conn.socket, sizeof(conn.socket) ) >= 0 ) {
		errno = EOK;
	    }
	    switch( errno ) {
	    case EISCONN:
	    case EALREADY:
	        errno = EOK;
		/* fall through */
	    case EOK:
		/* connect happened right away! */
		break;
	    case EINPROGRESS:
		break;
	    default:
		report_error( "connect" );
	    }
	    FD_SET( conn.h, &wfd );
	    conn.state = WAITING;
	    if( errno != EOK ) break;
	    /* fall through */
	case WAITING:
	    len = sizeof( status );
	    getsockopt( conn.h, SOL_SOCKET, SO_ERROR, &status, &len );
	    if( status == EISCONN || status == EOK ) {
		/* good - what we want */
	    } else if( status == ETIMEDOUT || status == ECONNREFUSED ) {
		conn.state = UNCREATED;
		/* wait a while and then try again */
		break;
	    } else if( status == EALREADY || status == EINPROGRESS ) {
		FD_SET( conn.h, &wfd );
		/* wait a while and then try again */
		break;
	    } else {
		report_error( "connection" );
	    }
	    config_socket( conn.h );
	    FD_SET( conn.h, &wfd );
	    conn.state = CONNECTED;
	    Trace0b( _TRACE_OUT_CONN, _TRACE_TESTING,
			    conn.socket.sin_len,
			    &conn.socket.sin_addr );
	    /* fall through */
	case CONNECTED:
	    /* check for input */
	    if( LEFT( conn.in ) > 0 ) {
		FD_SET( conn.h, &rfd );
		errno = EOK;
		len = read( conn.h, &conn.in.buff[conn.in.used], LEFT( conn.in ) );
		if( len < 0 ) {
		    switch( errno ) {
		    case EWOULDBLOCK:
		        len = 0;
			break;
		    case ENOTCONN:
		    case EPIPE:
		    case EINTR:
		    case ETIMEDOUT:
		    case ENETRESET:
		    case ECONNABORTED:
		    case ECONNRESET:
			broken( 0 );
			continue; /* back to top */
		    default:
			report_error( "read" );
		    }
		} else if( len == 0 && errno != EWOULDBLOCK ) {
		    broken( 0 );
		    continue; /* back to top */
		}
		conn.in.used += len;
	    }
	    if( conn.in.used > conn.in.min && conn.in.proxy ) {
		Trigger(conn.in.proxy);
		conn.in.proxy = 0;
	    }
	    /* check for output */
	    len = conn.out.used;
	    if( len > 0 ) {
		FD_SET( conn.h, &wfd );
		if( len > max_write ) len = max_write;
		len = write( conn.h, &conn.out.buff[0], len );
		if( len < 0 ) {
		    switch( errno ) {
		    case EWOULDBLOCK:
		        len = 0;
		        break;
		    case EINTR:
		    case EPIPE:
		    case ETIMEDOUT:
		    case ENETRESET:
		    case ECONNABORTED:
		    case ECONNRESET:
		    case ENOTCONN:
			broken( 0 );
			continue; /* back to the top */
		    default:
			report_error( "write" );
			break;
		    }
		}
		conn.out.used -= len;
		if( len != 0 && conn.out.used != 0 ) {
		    memmove( &conn.out.buff[0], &conn.out.buff[len], conn.out.used );
		}
	    }
	    if( LEFT( conn.out ) >= conn.out.min && conn.out.proxy != 0 ) {
		Trigger( conn.out.proxy );
		conn.out.proxy = 0;
	    }
	    break;
	}
	FD_SET( listen_socket, &rfd );
	if( select( FD_SETSIZE, &rfd, &wfd, NULL, &timeout ) < 0 ) {
	    if( errno != EINTR ) {
		report_error( "select" );
	    }
	}
    }
}


static void nonblock( int n )
{
    int		noblock;

    /* make connection nonblocking */
    noblock = 1;
    if( ioctl( conn.h, FIONBIO, &noblock) < 0 ) {
	report_error( "ioctl" );
    }
}

static void config_socket( int h )
{
    struct protoent	*proto;
    int			option;
    int ret;

    option = 1;
    proto = getprotobyname( "tcp" );
    ret=setsockopt( h, proto->p_proto, TCP_NODELAY, (void *)&option, sizeof( option ) );
    printf("ret=%d\n",ret);
    option = 1;
    ret=setsockopt( h, SOL_SOCKET, SO_KEEPALIVE, (void *)&option, sizeof( option ) );
    printf("ret=%d\n",ret);
    ret=setsockopt( h, SOL_SOCKET, SO_SNDLOWAT,  (void *)&option, 16 );
    printf("ret=%d\n",ret);
}


pid_t _select_receive( pid_t proxy )
{
    pid_t		pid;
    int         	nbytes;
    struct _mxfer_entry mx_entry[2];
    msg_t               type;
    void		*cb;

    for (;;) {
	pid  = Receive( 0, &msg, sizeof(msg) );
	if( pid == -1 || pid == proxy ) return( pid );

	type = msg.type;
	msg.status = ENOSYS;    /* if not match in switch below */
	nbytes = sizeof(msg.status);

	switch (type) {

	/*
	 *  Transmit packet functions
	 */
	case _DEV_ARM:
	    /*
	     *  Net.fd specs _DEV_EVENT_OUTPUT, which means that  
	     *  we should trigger the proxy when there is room
	     *  to tx a full-length packet
	     */
	    cb = (void *) __get_fd(pid, msg.dev_arm.fd, fd_ctrl);
	    if (cb == (void *) 0 || cb == (void *) - 1) {
		msg.status = EBADF;
		break;
	    }

	    if( LEFT( conn.out ) >= conn.out.min ) {
		/* if we have lots of buffers ... */
		Trigger(msg.dev_arm.proxy);
	    } else {
		/* else if we're low on buffers ... */
		conn.out.proxy = msg.dev_arm.proxy;
	    }
	    msg.dev_arm_reply.status = EOK;
	    Reply(pid, &msg.dev_arm_reply, sizeof(msg.dev_arm_reply));
	    nbytes = 0;
	    break;

	case _IO_WRITE:
        printf("got IO_WRITE\n");
	    cb = (void *) __get_fd(pid, msg.write.fd, fd_ctrl);
	    if (cb == (void *) 0 || cb == (void *) - 1) {
		msg.status = EBADF;
		break;
	    }
	    if( conn.out.error != EOK ) {
		msg.status = conn.out.error;
		conn.out.error = EOK;
		break;
	    }
	    nbytes = msg.write.nbytes;
/*
 *          Net.fd does two writes: one for it's header (see fd_hdr)
 *          followed immediately by the real ethernet-style packet,
 *          which Net.fd tacks a four-byte crc onto the end of.
 *
 *          You will probably want to store the fd_hdr somewhere
 *          associated with the cb, then when the second write
 *          comes in, tx the whole thing on your media as follows:
 *
 *          allocate packet buffer
 *          copy nbytes of data from msg.write.data to packet buffer
 *          pass buffer to hardware for tx
 */
#ifdef DEBUG
	    { 
		int I;
		qnx_display_msg("\nServer: tx: len: ");
		qnx_display_hex(nbytes);
		qnx_display_msg("data: ");
		for(I=0; I<min(20, nbytes); I+=2) {
		    qnx_display_hex(msg.write.data[I]<<8 | msg.write.data[I+1]);
		}
	    }
#endif
	    if( nbytes > LEFT( conn.out ) ) nbytes = LEFT( conn.out );
	    Readmsg( pid, sizeof( msg.write ) - sizeof( msg.write.data ),
	    	&conn.out.buff[conn.out.used], nbytes );
	    conn.out.used += nbytes;

	    msg.write_reply.status = EOK;
	    msg.write_reply.nbytes = nbytes;
	    Reply(pid, &msg.write_reply, sizeof(msg.write_reply));
	    nbytes = 0;

	    switch( conn.state ) {
	    case CONNECTED:
	        if( conn.out.used <= sizeof( struct fd_hdr_struct ) ) break;
		/* fall through */
	    case UNCONNECTED:
	        /*
		   get select to kick out and tell the loop in main that
		   there's data to write
		*/
	        errno = -1;
		return( -1 );
	    }
	    break;

	/*
	 *  Receive packet functions
	 */
	case _DEV_READ:
	    /*
	     *  Net.fd never asks for data with dev_read()
	     *  it passes a proxy to trigger when a mininum
	     *  number of bytes have arrived. 
	     */
	    cb = (void *) __get_fd(pid, msg.dev_read.fd, fd_ctrl);
	    if (cb == (void *) 0 || cb == (void *) - 1) {
		msg.status = EBADF;
		break;
	    }
	    conn.in.min = msg.dev_read.minimum;
	    if( conn.in.used >= conn.in.min ) {
		Trigger( msg.dev_read.proxy );
	    } else {
		conn.in.proxy = msg.dev_read.proxy;
	    }

	    msg.dev_read_reply.status  = EOK;
	    msg.dev_read_reply.nbytes  = 0; 
	    msg.dev_read_reply.timeout = 0; 
	    msg.dev_read_reply.empty   = 0;
	    Reply(pid, &msg.dev_read_reply, sizeof(msg.dev_read_reply));
	    nbytes = 0;
	    break;

	case _IO_READ:
        printf("got IO_READ\n");
	    cb = (void *) __get_fd(pid, msg.read.fd, fd_ctrl);
	    if (cb == (void *) 0 || cb == (void *) - 1) {
		msg.status = EBADF;
		break;
	    }
	    if( conn.in.error != EOK ) {
		msg.status = conn.in.error;
		conn.in.error = EOK;
		break;
	    }
	    nbytes = msg.read.nbytes;
	    if( conn.in.used < nbytes ) nbytes = conn.in.used;

	    _setmx(&mx_entry[1], &conn.in.buff[0], nbytes);

	    msg.read_reply.status = EOK;
	    msg.read_reply.nbytes = nbytes;
	    msg.read_reply.zero   = 0;
	    _setmx(&mx_entry[0], &msg,
	      sizeof(struct _io_read_reply) - sizeof(msg.read_reply.data));
	    Replymx(pid, 2, &mx_entry);

	    conn.in.used -= nbytes;
	    if( conn.in.used > 0 ) {
		memmove( &conn.in.buff[0], &conn.in.buff[nbytes], conn.in.used );
	    }
	    /*
	    	Kick out so that we have a chance to re-enable select
		on the read line (our input buffer might have been full).
	    */
	    errno = -1;
	    return( -1 );

	/*
	 *  Obviously, we must support open() and close().
	 *
	 *  Note that each Net.fd opens its device once, for
	 *  read/write, and will close when shutting down.
	 */
	case _IO_OPEN:
	    /* could look at mode - Net.fd only opens read/write */

#if 0 // caused crash, dunno why.
	    __get_pid_info(pid, &psdata3, fd_ctrl);
	    if (psdata3.euid != 0) {    /* gotta be root */
		msg.status = EPERM;
		break;
	    }
#endif

	    if (msg.open.path[0] != '\0') {
		msg.status = ENOENT;
		break;
	    }
	    if (!(cb = (void *) malloc(sizeof(char)))) {
		msg.status = ENOMEM;
		break;
	    }
	    if (qnx_fd_attach(pid, msg.open.fd,
		    0, 0, 0, 0, (unsigned) cb) == -1) {
		free(cb);
		msg.status = errno;
		break;
	    }
	    msg.status = EOK;
	    break;

	case _IO_CLOSE:
	    cb = (void *) __get_fd(pid, msg.close.fd, fd_ctrl);
	    if (cb == (void *) 0 || cb == (void *) - 1) {
		msg.status = EBADF;
		break;
	    }
	    free(cb);   /* Net.fd only opens once, so any close is final */
	    qnx_fd_attach(pid, msg.close.fd, 0, 0, 0, 0, 0);
	    msg.status = EOK;
	    break;

	/*
	 *  The following message are supported simply to
	 *  keep Net.fd happy, since Net.fd has to worry
	 *  about byte-stream considerations when using
	 *  /dev/ser.  For simplicity, this sample server
	 *  is assumed to talk to it's hardware (or whatever
	 *  is underneath it) on a packet basis.  If you wish
	 *  to support some kind of serial device, it might
	 *  be wise in that case to write a real dev driver
	 *  (like Dev.ser) rather than trying to duplicate Dev.
	 */
	case _DEV_TCGETATTR:
	    /* What's our media rate? */
	    msg.dev_tcgetattr_reply.status = EOK;
	    memset(&msg.dev_tcgetattr_reply.state, 0,
	      sizeof(msg.dev_tcgetattr_reply.state));
	    msg.dev_tcgetattr_reply.state.c_ispeed = 9600;
	    msg.dev_tcgetattr_reply.state.c_ospeed = 9600; /* Net.fd media rate */
	    Reply(pid, &msg.dev_tcgetattr_reply, sizeof(msg.dev_tcgetattr_reply));
	    nbytes = 0;
	    break;

	case _DEV_TCFLUSH:
	    /* wipe buffers.  That's nice */
	    conn.in.used = 0;
	    conn.out.used = 0;
	    msg.dev_tcflush_reply.status = EOK;
	    Reply(pid, &msg.dev_tcflush_reply, sizeof(msg.dev_tcflush_reply));
	    nbytes = 0;
	    break;

	case _DEV_MODE:
	    /* goto raw mode.  Sure. */
	    msg.dev_mode_reply.status = EOK;
	    Reply(pid, &msg.dev_mode_reply, sizeof(msg.dev_mode_reply));
	    nbytes = 0;
	    break;

	case _DEV_TCSETATTR:
	    /* ignore break, parity.  Check's in the mail. */
	    msg.dev_tcsetattr_reply.status = EOK;
	    Reply(pid, &msg.dev_tcsetattr_reply, sizeof(msg.dev_tcsetattr_reply));
	    nbytes = 0;
	    break;

	case _DEV_STATE:
	    /* clear events.  Why not. */
	    msg.dev_state_reply.status = EOK;
	    Reply(pid, &msg.dev_state_reply, sizeof(msg.dev_state_reply));
	    nbytes = 0;
	    break;

	case _DEV_OSIZE:
	    /* Don't trigger Net.fd for output until 1600 bytes free */
	    if( msg.dev_osize.size != -1 ) {
		conn.out.min = msg.dev_osize.size;
	    }
	    msg.dev_osize_reply.status = EOK;
	    Reply(pid, &msg.dev_osize_reply, sizeof(msg.dev_osize_reply));
	    nbytes = 0;
	    break;

	case _DEV_TCSETCT:
	    /* Net.fd wants hangups.  That's nice. */
	    msg.dev_tcsetct_reply.status = EOK;
	    Reply(pid, &msg.dev_tcsetct_reply, sizeof(msg.dev_tcsetct_reply));
	    nbytes = 0;
	    break;

	/*
	 *  make ls work
	 */
	case _IO_STAT:
	    if (msg.open.path[0] != '\0') {
		msg.status = ENOENT;
		break;
	    }
	    Reply(pid, &fstat_reply, sizeof(fstat_reply));
	    nbytes = 0;
	    break;

	/*
	 *  make sin ver work
	 */
	case _SYSMSG:
	    switch (msg.sysmsg.subtype) {
	    case _SYSMSG_SUBTYPE_VERSION:
		msg.version_reply.status = EOK;
		_setmx(&mx_entry[0], &msg, sizeof(msg.version_reply));
		_setmx(&mx_entry[1], &version_stamp, sizeof(version_stamp));
		Replymx(pid, 2, &mx_entry);
		nbytes = 0;
		break;
	    }
	    break;

#ifdef DEBUG
	default:
	    qnx_display_msg("\nServer: unknown msg type = ");
	    qnx_display_hex(type);
#endif
	}

	if (nbytes) {
	    Reply(pid, &msg, nbytes);
	}
    }
}
